.nolist
include \masm32\include\masm32rt.inc
.586
.xmm
include \masm32\macros\timers.asm		; search for SLEEP_MS to find the start of the tests

	LOOP_COUNT	= 100
	SLEEP_MS	= 200
	SLOT_COUNT	= 10*1000*1000

findslotSSE2	PROTO:DWORD
findslotScasd	PROTO:DWORD
findslotMW	PROTO:DWORD

ShowCPU	PROTO:DWORD
  
AlgoTest MACRO txt, pAlgo, arg		; ########## THE MACRO THAT PERFORMS THE TESTS ##########
  % print <txt>
	counter_begin LOOP_COUNT, HIGH_PRIORITY_CLASS
		invoke pAlgo, arg
		mov esi, eax
	counter_end
	shr eax, 10
	print ustr$(eax),9, "kilocycles, result: slot "
	print str$(esi), 13, 10
ENDM

CodeSize MACRO arg
	tmp$ CATSTR <print " >, <arg>, <", 9>
	% echo tmp$
	tmp$
	tmp$ CATSTR <mov eax, offset >, <arg>, <e->, <arg>
	% echo tmp$
	tmp$
	print str$(eax), 13, 10
ENDM

.code
; 
start:
	invoke ShowCPU, 1				; ***** show which CPU you are running ****
	print chr$(13, 10)
	mov edi, alloc$(SLOT_COUNT*4+32)
	memalign edi, 16
	push edi
	mov eax, 1234abcdh				; fill slots
	m2m ecx, SLOT_COUNT
	.Repeat
		stosd
		dec ecx
	.Until Zero?
	pop edi

	print chr$(13, 10, "Code sizes: ", 13, 10)
	CodeSize findslotSSE2
	CodeSize findslotMW
	CodeSize findslotScasd

	print chr$(13, 10, "Timings: ", 13, 10)

	mov dword ptr [edi+SLOT_COUNT*4], 0	; put a zero into the slot we want to find
	invoke Sleep, SLEEP_MS
	AlgoTest " findslotSSE2   ", findslotSSE2, edi
	AlgoTest " findslotMW     ", findslotMW, edi
	AlgoTest " findslotScasd  ", findslotScasd, edi

	print chr$(13, 10)
	mov dword ptr [edi+SLOT_COUNT*4/10], 0	; put a zero into the slot we want to find
	invoke Sleep, SLEEP_MS
	AlgoTest " findslotSSE2   ", findslotSSE2, edi
	AlgoTest " findslotMW     ", findslotMW, edi
	AlgoTest " findslotScasd  ", findslotScasd, edi

	print chr$(13, 10)
	mov dword ptr [edi+SLOT_COUNT*4/100], 0	; put a zero into the slot we want to find
	invoke Sleep, SLEEP_MS
	AlgoTest " findslotSSE2   ", findslotSSE2, edi
	AlgoTest " findslotMW     ", findslotMW, edi
	AlgoTest " findslotScasd  ", findslotScasd, edi

    	inkey chr$(13, 10, 9, 9, 9, 9, 9, "-- hit any key --")
    exit

;  ALL ALGOS ARE WITHOUT STACK FRAME!! 
OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

align 16
findslotScasd proc src:DWORD
	xor eax, eax
	push edi
	mov edi, [esp+8]
	or ecx, -1
	; int 3
	repne scasd
	lea eax, [edi-4]
	sub eax, [esp+8]
	shr eax, 2
	pop edi
	ret 4
findslotScasd endp
findslotScasde:

align 16
findslotSSE2 proc src:DWORD
	mov eax, [esp+4]
	pxor xmm0, xmm0			; zero for comparison
	pxor xmm1, xmm1			; zero for comparison
	push ecx					; all registers preserved, except eax = return value
	push edx
@@:	pcmpeqd xmm0, [eax]
	pcmpeqd xmm1, [eax+16]	; compare packed bytes in [m128] and xmm1 for equality
	lea eax, [eax+32]			; len counter (moving up lea or add costs 3 cycles for the 191 byte string)
	por xmm1, xmm0			; or them: one of the mem locations may contain a nullbyte
	pmovmskb edx, xmm1	; set byte mask in edx
	test edx, edx
	jz @B

	sub eax, [esp+4+8]	; subtract original src pointer
	shl edx, 16				; create space for the ecx bytes
	pmovmskb ecx, xmm0	; set byte mask in ecx (has to be repeated, sorry)
	or edx, ecx				; combine xmm0 and xmm1 results
	bsf edx, edx			; bit scan for the index
	lea eax, [edx+eax-32]
	shr eax, 2
	pop edx
	pop ecx
	ret 4
findslotSSE2 endp
findslotSSE2e:

align 16
findslotMW proc base:DWORD

    mov edx, [esp+4]
    mov eax, -2
    jmp @F
    align 16
  @@:
    add eax, 2
    mov ecx, -1
    and ecx, [edx+eax*4]
    test ecx, [edx+eax*4+4]
    jnz @B
    test DWORD PTR [edx+eax*4], -1
    jz  @F
    add eax, 1
  @@:
    ret 4
findslotMW endp
findslotMWe:

; ######################### END OF STRLEN ALGOS ##################################

.data?
	osv		OSVERSIONINFO <?>
	CpuFlags	dd ?
	CpuExFlags	dd ?
	CpuMMX		dd ?
	CpuSSE		dd ?
	CpuFxSR		dd ?
	CpuManu		dd 20 dup (?)
	CpuInfo		dd 100 dup (?)
	CpuBrand	dd 13 dup (?)		; 48 bytes
	step	dd ?
	model	dd ?
	family	dd ?
	ptype	dd ?
	extmodel	dd ?
	extfamily	dd ?
	CpuAMD	dd ?
	CpuIntel	dd ?
.data
	MbCopyDestAlign	dd -1			; default
	txModel	db "Model        ", 9, 0
	txFam	db 13, 10, "Family       ", 9, 0
	txStep	db 13, 10, "Step           ", 9, 0
	txType	db 13, 10, "Type          ", 9, 0
	txExtM	db 13, 10, "Ext model   ", 9, 0
	txExtF	db 13, 10, "Ext family   ", 9, 0
	txCpuMan	db 13, 10, "Manufacturer", 9, 0
	txDesc	db 13, 10, "Description", 9, 0
	txBrand	db 13, 10, "Brand name", 9, 0
	txP1	db "P1 (1993+)", 0
	txPro	db "Pentium Pro (1995+)", 0
	txP2	db "P2 (1997+)", 0
	txP3	db "P3 (2000+)", 0
	txP4	db "P4 (2000+)", 0
	txIta1	db "Itanium 1 (2001+)", 0
	txIta2	db "Itanium 2 (2002+)", 0
	txPM	db "Pentium M (2003+)", 0
	txCore	db "Core (2006+)", 0
	txCore2	db "Core Duo (2006+)", 0
	txP4Prescott	db "Pentium 4 Prescott (2005+)", 0
	txUnknownP4	db "Pentium 4 (?)", 0
	txUnknownAMD	db "AMD (?)", 0
	txUnknown	db "?", 0
.code

ShowCPU proc mode:DWORD
; Usage:

; push 0
; call ShowCPU 	; for info only

; push -1
; call ShowCPU	; for printing the long version

; push 1
; call ShowCPU	; for printing the short version

  push edi
  push esi
  push ebx
  xor eax, eax
  db 0Fh, 0A2h		; cpuid 0
  mov CpuManu, ebx
  mov CpuManu[4], edx
  mov CpuManu[8], ecx
  .if ecx=="DMAc"
	inc CpuAMD				; it's an AMD
  .elseif ecx=="letn"
  	inc CpuIntel				; it's an Intel
  .endif
  mov esi, 80000002h
  mov edi, offset CpuBrand
  .Repeat
	mov eax, esi
	db 0Fh, 0A2h			; cpuid 80000002h-80000004h
	stosd
	mov eax, ebx
	stosd
	mov eax, ecx
	stosd
	mov eax, edx
	stosd
	inc esi
  .Until esi==80000005h
;  MsgBox 0, offset CpuBrand, "Brand", MB_OK
  push 1
  pop eax
  db 0Fh, 0A2h					; cpuid 1
  mov CpuFlags, edx
  mov CpuExFlags, ecx
  and CpuMMX, 0
  and CpuSSE, 0

  mov ebx, eax
  and eax, 15
  mov step, eax

  shr ebx, 4
  mov eax, ebx
  and eax, 15
  mov model, eax

  shr ebx, 4
  mov eax, ebx
  and eax, 15
  mov family, eax

  shr ebx, 4
  mov eax, ebx
  and eax, 15
  mov ptype, eax

  shr ebx, 4
  mov eax, ebx
  and eax, 15
  mov extmodel, eax

  shr ebx, 4
  mov eax, ebx
  and eax, 15
  mov extfamily, eax
  mov edi, offset CpuInfo
  and CpuInfo, 0
  mov eax, [esp+4+12]	; get arg from stack (we pushed ebx)
  .if eax==-1
;	invoke lstrcat, edi, chr$("CPU family ")
;	invoke lstrcat, edi, str$(family)
;	invoke lstrcat, edi, chr$(", model ")
;	invoke lstrcat, edi, str$(model)
;	invoke lstrcat, edi, ", "
 
	invoke lstrcat, edi, offset txModel
	invoke lstrcat, edi, str$(model)

	invoke lstrcat, edi, offset txFam
	invoke lstrcat, edi, str$(family)

	.if step
	invoke lstrcat, edi, offset txStep
	invoke lstrcat, edi, str$(step)
	.endif

	.if ptype
	invoke lstrcat, edi, offset txType
	invoke lstrcat, edi, str$(ptype)
	.endif

	.if extmodel
	invoke lstrcat, edi, offset txExtM
	invoke lstrcat, edi, str$(extmodel)
	.endif

	.if extfamily
	invoke lstrcat, edi, offset txExtF
	invoke lstrcat, edi, str$(extfamily)
	.endif

	invoke lstrcat, edi, offset txCpuMan
	invoke lstrcat, edi, offset CpuManu

	invoke lstrcat, edi, offset txDesc
  .endif
  .if CpuAMD
	invoke lstrcat, edi, chr$("AMD ")
	; and MbCopyDestAlign, 0	; in case AMD runs better on source data alignment
  .elseif CpuIntel
	invoke lstrcat, edi, chr$("Intel ")
  .endif
  .if family==5
	mov edx, offset txP1
  .elseif family==6
	.if model<=1
		mov edx, offset txPro
	.elseif model==3 || model==5 || model==6
		mov edx, offset txP2
	.elseif model==7 || model==8 || model==10 || model==11
		mov edx, offset txP3
	.elseif model==9 || model==13
		mov edx, offset txPM
	.elseif model==14
		mov edx, offset txCore
	.elseif model==15
		mov edx, offset txCore2
	.endif
  .elseif family==7	; Itanium
		mov edx, offset txIta1
  .elseif family==15
	.if model==0
		mov edx, offset txP4
	.elseif model<=2
		mov edx, offset txIta2
	.elseif model==4
		mov edx, offset txP4Prescott
	.elseif model==6
		mov edx, chr$("Pentium dual-core x64")
	.elseif model==11 && CpuAMD
		mov edx, chr$("AMD XP dual-core x64")
	.elseif CpuAMD
		mov edx, offset txUnknownAMD
	.else
		mov edx, offset txUnknownP4
	.endif
  .elseif CpuAMD
	mov edx, offset txUnknownAMD
  .else
	mov edx, offset txUnknown
  .endif
  invoke lstrcat, edi, edx
  mov edx, CpuFlags
  bt edx, 25
  jnc @F
  inc CpuSSE	; edx bit 25, SSE1
@@:

  bt edx, 26	; edx bit 26, SSE2
  jnc @F
  inc CpuSSE
@@:

  mov ecx, CpuExFlags
  bt ecx, 0		; ecx bit 0, SSE3
  jnc @F
  inc CpuSSE
@@:
  bt ecx, 9		; ecx bit 9, SSE4
  jnc @F
  inc CpuSSE
@@:
  bt edx, 23
  .if Carry?
	inc CpuMMX
	.if CpuSSE==0
		invoke lstrcat, edi, chr$(", MMX")
	.endif
  .else
	invoke lstrcat, edi, chr$(", no MMX")
  .endif
  mov edx, CpuFlags
  bt edx, 24
  jnc @F
  inc CpuFxSR	; supports FXSAVE, FXRSTOR
@@:
  .if CpuSSE
	invoke lstrcat, edi, chr$(", SSE")
	invoke lstrcat, edi, str$(CpuSSE)
  .endif
  invoke lstrcat, edi, offset txBrand
  invoke lstrcat, edi, offset CpuBrand
  mov eax, [esp+4+12]
  .if eax==-1
	print edi, 13, 10
  .elseif eax==1
	print offset CpuBrand
	.if CpuSSE
		print chr$(32, 40, "SSE")
		print str$(CpuSSE), 41
	.endif
  .endif
  pop ebx
  pop esi
  pop edi
  ret 4
ShowCPU endp

end start

OPT_Assembler	ml
OPT_Res	0
OPT_Tmp2Asm	1
OPT_Wait	0

Timings:
AMD Athlon(tm) 64 X2 Dual Core Processor 4000+ (SSE3)
findslotSSE2   32452 kilocycles, result: slot 10000000
findslotScasd  40024 kilocycles, result: slot 10000000

findslotSSE2   3155 kilocycles, result: slot 1000000
findslotScasd  3905 kilocycles, result: slot 1000000

findslotSSE2   208 kilocycles, result: slot 100000
findslotScasd  300 kilocycles, result: slot 100000

Intel(R) Pentium(R) 4 CPU 3.40GHz (SSE3)
findslotSSE2   45095 kilocycles, result: slot 10000000
findslotScasd  47280 kilocycles, result: slot 10000000

findslotSSE2   4536 kilocycles, result: slot 1000000
findslotScasd  4712 kilocycles, result: slot 1000000

findslotSSE2   162 kilocycles, result: slot 100000
findslotScasd  395 kilocycles, result: slot 100000

Intel(R) Pentium(R) 4 CPU 2.40GHz (SSE2)
 findslotSSE2   50004 kilocycles, result: slot 10000000
 findslotScasd  54834 kilocycles, result: slot 10000000

 findslotSSE2   5014 kilocycles, result: slot 1000000
 findslotScasd  5471 kilocycles, result: slot 1000000

 findslotSSE2   103 kilocycles, result: slot 100000
 findslotScasd  397 kilocycles, result: slot 100000
 